=== modified file 'DBUS-API'
--- DBUS-API	2011-10-02 19:18:24 +0000
+++ DBUS-API	2011-11-24 19:09:29 +0000
@@ -152,6 +152,9 @@
 *** Rejected(s: Reason)
     This client was not given its secret for a specified Reason.
 
+*** NewRequest(s: IP address)
+    A client at IP adress has asked for its password.
+
 * Copyright
 
     Copyright © 2010-2011 Teddy Hogeborn

=== modified file 'Makefile'
--- Makefile	2011-10-15 16:48:03 +0000
+++ Makefile	2011-11-26 20:59:56 +0000
@@ -260,13 +260,8 @@
 
 # Run the server with a local config
 run-server: confdir/mandos.conf confdir/clients.conf
-	@echo "#################################################################"
-	@echo "# NOTE: Please IGNORE the error about \"Could not open file      #"
-	@echo "# u'/var/run/mandos.pid'\" -  it is harmless and is caused by    #"
-	@echo "# the server not running as root.  Do NOT run \"make run-server\" #"
-	@echo "# server as root if you didn't also unpack and compile it thus. #"
-	@echo "#################################################################"
-	./mandos --debug --no-dbus --configdir=confdir $(SERVERARGS)
+	./mandos --debug --no-dbus --configdir=confdir --no-restore \
+	$(SERVERARGS)
 
 # Used by run-server
 confdir/mandos.conf: mandos.conf

=== modified file 'TODO'
--- TODO	2011-10-14 18:00:50 +0000
+++ TODO	2011-11-09 17:16:03 +0000
@@ -41,8 +41,7 @@
 
 * mandos (server)
 ** TODO [#B] Log level							  :BUGS:
-** TODO Persistent state						  :BUGS:
-   /var/lib/mandos/*
+  
 *** TODO /etc/mandos/clients.d/*.conf
     Watch this directory and add/remove/update clients?
 ** TODO [#C] config for TXT record
@@ -53,8 +52,6 @@
    [[info:standards:Option%20Table][Table of Long Options]]
 ** TODO Implement --socket
    [[info:standards:Option%20Table][Table of Long Options]]
-** TODO Date+time on console log messages				  :BUGS:
-   Is this the default?
 ** TODO [#C] DBusServiceObjectUsingSuper
 ** TODO [#B] Global enable/disable flag
 ** TODO [#B] By-client countdown on secrets given
@@ -68,7 +65,6 @@
    http://0pointer.de/blog/projects/systemd.html
    http://wiki.debian.org/systemd
 ** TODO Separate logging logic to own object
-** TODO make clients to a dict!
 ** TODO [#A] Limit approval_delay to max gnutls/tls timeout value
 ** TODO [#B] break the wait on approval_delay if connection dies
 ** TODO Generate Client.runtime_expansions from client options + extra

=== modified file 'debian/control'
--- debian/control	2011-10-11 19:36:00 +0000
+++ debian/control	2011-11-26 20:59:56 +0000
@@ -17,7 +17,8 @@
 Architecture: all
 Depends: ${misc:Depends}, python (>=2.6), python-gnutls, python-dbus,
 	python-avahi, python-gobject, avahi-daemon, adduser,
-	python-urwid, python (>=2.7) | python-argparse
+	python-urwid, python (>=2.7) | python-argparse,
+	python-gnupginterface
 Recommends: fping
 Description: server giving encrypted passwords to Mandos clients
  This is the server part of the Mandos system, which allows

=== modified file 'mandos'
--- mandos	2011-10-15 16:48:03 +0000
+++ mandos	2011-11-26 20:59:56 +0000
@@ -63,6 +63,8 @@
 import cPickle as pickle
 import multiprocessing
 import types
+import binascii
+import tempfile
 
 import dbus
 import dbus.service
@@ -73,6 +75,7 @@
 import ctypes.util
 import xml.dom.minidom
 import inspect
+import GnuPGInterface
 
 try:
     SO_BINDTODEVICE = socket.SO_BINDTODEVICE
@@ -84,22 +87,125 @@
 
 
 version = "1.4.1"
+stored_state_path = "/var/lib/mandos/clients.pickle"
 
-#logger = logging.getLogger('mandos')
-logger = logging.Logger('mandos')
+logger = logging.getLogger()
 syslogger = (logging.handlers.SysLogHandler
              (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
               address = str("/dev/log")))
-syslogger.setFormatter(logging.Formatter
-                       ('Mandos [%(process)d]: %(levelname)s:'
-                        ' %(message)s'))
-logger.addHandler(syslogger)
-
-console = logging.StreamHandler()
-console.setFormatter(logging.Formatter('%(name)s [%(process)d]:'
-                                       ' %(levelname)s:'
-                                       ' %(message)s'))
-logger.addHandler(console)
+
+try:
+    if_nametoindex = (ctypes.cdll.LoadLibrary
+                      (ctypes.util.find_library("c"))
+                      .if_nametoindex)
+except (OSError, AttributeError):
+    def if_nametoindex(interface):
+        "Get an interface index the hard way, i.e. using fcntl()"
+        SIOCGIFINDEX = 0x8933  # From /usr/include/linux/sockios.h
+        with contextlib.closing(socket.socket()) as s:
+            ifreq = fcntl.ioctl(s, SIOCGIFINDEX,
+                                struct.pack(str("16s16x"),
+                                            interface))
+        interface_index = struct.unpack(str("I"),
+                                        ifreq[16:20])[0]
+        return interface_index
+
+
+def initlogger(level=logging.WARNING):
+    """init logger and add loglevel"""
+    
+    syslogger.setFormatter(logging.Formatter
+                           ('Mandos [%(process)d]: %(levelname)s:'
+                            ' %(message)s'))
+    logger.addHandler(syslogger)
+    
+    console = logging.StreamHandler()
+    console.setFormatter(logging.Formatter('%(asctime)s %(name)s'
+                                           ' [%(process)d]:'
+                                           ' %(levelname)s:'
+                                           ' %(message)s'))
+    logger.addHandler(console)
+    logger.setLevel(level)
+
+
+class CryptoError(Exception):
+    pass
+
+
+class Crypto(object):
+    """A simple class for OpenPGP symmetric encryption & decryption"""
+    def __init__(self):
+        self.gnupg = GnuPGInterface.GnuPG()
+        self.tempdir = tempfile.mkdtemp(prefix="mandos-")
+        self.gnupg = GnuPGInterface.GnuPG()
+        self.gnupg.options.meta_interactive = False
+        self.gnupg.options.homedir = self.tempdir
+        self.gnupg.options.extra_args.extend(['--force-mdc',
+                                              '--quiet'])
+    
+    def __enter__(self):
+        return self
+    
+    def __exit__ (self, exc_type, exc_value, traceback):
+        self._cleanup()
+        return False
+    
+    def __del__(self):
+        self._cleanup()
+    
+    def _cleanup(self):
+        if self.tempdir is not None:
+            # Delete contents of tempdir
+            for root, dirs, files in os.walk(self.tempdir,
+                                             topdown = False):
+                for filename in files:
+                    os.remove(os.path.join(root, filename))
+                for dirname in dirs:
+                    os.rmdir(os.path.join(root, dirname))
+            # Remove tempdir
+            os.rmdir(self.tempdir)
+            self.tempdir = None
+    
+    def password_encode(self, password):
+        # Passphrase can not be empty and can not contain newlines or
+        # NUL bytes.  So we prefix it and hex encode it.
+        return b"mandos" + binascii.hexlify(password)
+    
+    def encrypt(self, data, password):
+        self.gnupg.passphrase = self.password_encode(password)
+        with open(os.devnull) as devnull:
+            try:
+                proc = self.gnupg.run(['--symmetric'],
+                                      create_fhs=['stdin', 'stdout'],
+                                      attach_fhs={'stderr': devnull})
+                with contextlib.closing(proc.handles['stdin']) as f:
+                    f.write(data)
+                with contextlib.closing(proc.handles['stdout']) as f:
+                    ciphertext = f.read()
+                proc.wait()
+            except IOError as e:
+                raise CryptoError(e)
+        self.gnupg.passphrase = None
+        return ciphertext
+    
+    def decrypt(self, data, password):
+        self.gnupg.passphrase = self.password_encode(password)
+        with open(os.devnull) as devnull:
+            try:
+                proc = self.gnupg.run(['--decrypt'],
+                                      create_fhs=['stdin', 'stdout'],
+                                      attach_fhs={'stderr': devnull})
+                with contextlib.closing(proc.handles['stdin'] ) as f:
+                    f.write(data)
+                with contextlib.closing(proc.handles['stdout']) as f:
+                    decrypted_plaintext = f.read()
+                proc.wait()
+            except IOError as e:
+                raise CryptoError(e)
+        self.gnupg.passphrase = None
+        return decrypted_plaintext
+
+
 
 class AvahiError(Exception):
     def __init__(self, value, *args, **kwargs):
@@ -164,10 +270,6 @@
                             .GetAlternativeServiceName(self.name))
         logger.info("Changing Zeroconf service name to %r ...",
                     self.name)
-        syslogger.setFormatter(logging.Formatter
-                               ('Mandos (%s) [%%(process)d]:'
-                                ' %%(levelname)s: %%(message)s'
-                                % self.name))
         self.remove()
         try:
             self.add()
@@ -193,7 +295,7 @@
                 avahi.DBUS_INTERFACE_ENTRY_GROUP)
         self.entry_group_state_changed_match = (
             self.group.connect_to_signal(
-                'StateChanged', self .entry_group_state_changed))
+                'StateChanged', self.entry_group_state_changed))
         logger.debug("Adding Zeroconf service '%s' of type '%s' ...",
                      self.name, self.type)
         self.group.AddService(
@@ -225,7 +327,7 @@
             try:
                 self.group.Free()
             except (dbus.exceptions.UnknownMethodException,
-                    dbus.exceptions.DBusException) as e:
+                    dbus.exceptions.DBusException):
                 pass
             self.group = None
         self.remove()
@@ -265,6 +367,15 @@
                                  self.server_state_changed)
         self.server_state_changed(self.server.GetState())
 
+class AvahiServiceToSyslog(AvahiService):
+    def rename(self):
+        """Add the new name to the syslog messages"""
+        ret = AvahiService.rename(self)
+        syslogger.setFormatter(logging.Formatter
+                               ('Mandos (%s) [%%(process)d]:'
+                                ' %%(levelname)s: %%(message)s'
+                                % self.name))
+        return ret
 
 def _timedelta_to_milliseconds(td):
     "Convert a datetime.timedelta() to milliseconds"
@@ -289,8 +400,9 @@
                      instance %(name)s can be used in the command.
     checker_initiator_tag: a gobject event source tag, or None
     created:    datetime.datetime(); (UTC) object creation
+    client_structure: Object describing what attributes a client has
+                      and is used for storing the client at exit
     current_checker_command: string; current running checker_command
-    disable_hook:  If set, called by disable() as disable_hook(self)
     disable_initiator_tag: a gobject event source tag, or None
     enabled:    bool()
     fingerprint: string (40 or 32 hexadecimal digits); used to
@@ -299,6 +411,10 @@
     interval:   datetime.timedelta(); How often to start a new checker
     last_approval_request: datetime.datetime(); (UTC) or None
     last_checked_ok: datetime.datetime(); (UTC) or None
+
+    last_checker_status: integer between 0 and 255 reflecting exit
+                         status of last checker. -1 reflects crashed
+                         checker, or None.
     last_enabled: datetime.datetime(); (UTC)
     name:       string; from the config file, used in log messages and
                         D-Bus identifiers
@@ -331,7 +447,7 @@
     def approval_delay_milliseconds(self):
         return _timedelta_to_milliseconds(self.approval_delay)
     
-    def __init__(self, name = None, disable_hook=None, config=None):
+    def __init__(self, name = None, config=None):
         """Note: the 'checker' key in 'config' sets the
         'checker_command' attribute and *not* the 'checker'
         attribute."""
@@ -357,23 +473,22 @@
                             % self.name)
         self.host = config.get("host", "")
         self.created = datetime.datetime.utcnow()
-        self.enabled = False
+        self.enabled = True
         self.last_approval_request = None
-        self.last_enabled = None
+        self.last_enabled = datetime.datetime.utcnow()
         self.last_checked_ok = None
+        self.last_checker_status = None
         self.timeout = string_to_delta(config["timeout"])
         self.extended_timeout = string_to_delta(config
                                                 ["extended_timeout"])
         self.interval = string_to_delta(config["interval"])
-        self.disable_hook = disable_hook
         self.checker = None
         self.checker_initiator_tag = None
         self.disable_initiator_tag = None
-        self.expires = None
+        self.expires = datetime.datetime.utcnow() + self.timeout
         self.checker_callback_tag = None
         self.checker_command = config["checker"]
         self.current_checker_command = None
-        self.last_connect = None
         self._approved = None
         self.approved_by_default = config.get("approved_by_default",
                                               True)
@@ -385,11 +500,22 @@
         self.changedstate = (multiprocessing_manager
                              .Condition(multiprocessing_manager
                                         .Lock()))
+        self.client_structure = [attr for attr in
+                                 self.__dict__.iterkeys()
+                                 if not attr.startswith("_")]
+        self.client_structure.append("client_structure")
+        
+        for name, t in inspect.getmembers(type(self),
+                                          lambda obj:
+                                              isinstance(obj,
+                                                         property)):
+            if not name.startswith("_"):
+                self.client_structure.append(name)
     
+    # Send notice to process children that client state has changed
     def send_changedstate(self):
-        self.changedstate.acquire()
-        self.changedstate.notify_all()
-        self.changedstate.release()
+        with self.changedstate:
+            self.changedstate.notify_all()
     
     def enable(self):
         """Start this client's checker and timeout hooks"""
@@ -397,20 +523,10 @@
             # Already enabled
             return
         self.send_changedstate()
-        # Schedule a new checker to be started an 'interval' from now,
-        # and every interval from then on.
-        self.checker_initiator_tag = (gobject.timeout_add
-                                      (self.interval_milliseconds(),
-                                       self.start_checker))
-        # Schedule a disable() when 'timeout' has passed
         self.expires = datetime.datetime.utcnow() + self.timeout
-        self.disable_initiator_tag = (gobject.timeout_add
-                                   (self.timeout_milliseconds(),
-                                    self.disable))
         self.enabled = True
         self.last_enabled = datetime.datetime.utcnow()
-        # Also start a new checker *right now*.
-        self.start_checker()
+        self.init_checker()
     
     def disable(self, quiet=True):
         """Disable this client."""
@@ -428,23 +544,33 @@
             gobject.source_remove(self.checker_initiator_tag)
             self.checker_initiator_tag = None
         self.stop_checker()
-        if self.disable_hook:
-            self.disable_hook(self)
         self.enabled = False
         # Do not run this again if called by a gobject.timeout_add
         return False
     
     def __del__(self):
-        self.disable_hook = None
         self.disable()
     
+    def init_checker(self):
+        # Schedule a new checker to be started an 'interval' from now,
+        # and every interval from then on.
+        self.checker_initiator_tag = (gobject.timeout_add
+                                      (self.interval_milliseconds(),
+                                       self.start_checker))
+        # Schedule a disable() when 'timeout' has passed
+        self.disable_initiator_tag = (gobject.timeout_add
+                                   (self.timeout_milliseconds(),
+                                    self.disable))
+        # Also start a new checker *right now*.
+        self.start_checker()
+    
     def checker_callback(self, pid, condition, command):
         """The checker has completed, so take appropriate actions."""
         self.checker_callback_tag = None
         self.checker = None
         if os.WIFEXITED(condition):
-            exitstatus = os.WEXITSTATUS(condition)
-            if exitstatus == 0:
+            self.last_checker_status =  os.WEXITSTATUS(condition)
+            if self.last_checker_status == 0:
                 logger.info("Checker for %(name)s succeeded",
                             vars(self))
                 self.checked_ok()
@@ -452,6 +578,7 @@
                 logger.info("Checker for %(name)s failed",
                             vars(self))
         else:
+            self.last_checker_status = -1
             logger.warning("Checker for %(name)s crashed?",
                            vars(self))
     
@@ -691,7 +818,7 @@
         
         Note: Will not include properties with access="write".
         """
-        all = {}
+        properties = {}
         for name, prop in self._get_all_dbus_properties():
             if (interface_name
                 and interface_name != prop._dbus_interface):
@@ -702,11 +829,11 @@
                 continue
             value = prop()
             if not hasattr(value, "variant_level"):
-                all[name] = value
+                properties[name] = value
                 continue
-            all[name] = type(value)(value, variant_level=
-                                    value.variant_level+1)
-        return dbus.Dictionary(all, signature="sv")
+            properties[name] = type(value)(value, variant_level=
+                                           value.variant_level+1)
+        return dbus.Dictionary(properties, signature="sv")
     
     @dbus.service.method(dbus.INTROSPECTABLE_IFACE,
                          out_signature="s",
@@ -763,6 +890,7 @@
     return dbus.String(dt.isoformat(),
                        variant_level=variant_level)
 
+
 class AlternateDBusNamesMetaclass(DBusObjectWithProperties
                                   .__metaclass__):
     """Applied to an empty subclass of a D-Bus object, this metaclass
@@ -860,6 +988,7 @@
                                         attribute.func_closure)))
         return type.__new__(mcs, name, bases, attr)
 
+
 class ClientDBus(Client, DBusObjectWithProperties):
     """A Client class using D-Bus
     
@@ -874,9 +1003,10 @@
     # dbus.service.Object doesn't use super(), so we can't either.
     
     def __init__(self, bus = None, *args, **kwargs):
+        self.bus = bus
+        Client.__init__(self, *args, **kwargs)
+        
         self._approvals_pending = 0
-        self.bus = bus
-        Client.__init__(self, *args, **kwargs)
         # Only now, when this client is initialized, can it show up on
         # the D-Bus
         client_object_name = unicode(self.name).translate(
@@ -892,7 +1022,7 @@
                              variant_level=1):
         """ Modify a variable so that it's a property which announces
         its changes to DBus.
-
+        
         transform_fun: Function that takes a value and a variant_level
                        and transforms it to a D-Bus type.
         dbus_name: D-Bus name of the variable
@@ -1052,6 +1182,14 @@
         "D-Bus signal"
         return self.need_approval()
     
+    # NeRwequest - signal
+    @dbus.service.signal(_interface, signature="s")
+    def NewRequest(self, ip):
+        """D-Bus signal
+        Is sent after a client request a password.
+        """
+        pass
+    
     ## Methods
     
     # Approve - method
@@ -1286,9 +1424,11 @@
             return super(ProxyClient, self).__setattr__(name, value)
         self._pipe.send(('setattr', name, value))
 
+
 class ClientDBusTransitional(ClientDBus):
     __metaclass__ = AlternateDBusNamesMetaclass
 
+
 class ClientHandler(socketserver.BaseRequestHandler, object):
     """A class to handle client connections.
     
@@ -1355,6 +1495,9 @@
                     logger.warning("Bad certificate: %s", error)
                     return
                 logger.debug("Fingerprint: %s", fpr)
+                if self.server.use_dbus:
+                    # Emit D-Bus signal
+                    client.NewRequest(str(self.client_address))
                 
                 try:
                     client = ProxyClient(child_pipe, fpr,
@@ -1504,7 +1647,7 @@
         # Convert the buffer to a Python bytestring
         fpr = ctypes.string_at(buf, buf_len.value)
         # Convert the bytestring to hexadecimal notation
-        hex_fpr = ''.join("%02X" % ord(char) for char in fpr)
+        hex_fpr = binascii.hexlify(fpr).upper()
         return hex_fpr
 
 
@@ -1624,7 +1767,7 @@
         self.enabled = False
         self.clients = clients
         if self.clients is None:
-            self.clients = set()
+            self.clients = {}
         self.use_dbus = use_dbus
         self.gnutls_priority = gnutls_priority
         IPv6_TCPServer.__init__(self, server_address,
@@ -1677,7 +1820,7 @@
             fpr = request[1]
             address = request[2]
             
-            for c in self.clients:
+            for c in self.clients.itervalues():
                 if c.fingerprint == fpr:
                     client = c
                     break
@@ -1767,30 +1910,6 @@
     return timevalue
 
 
-def if_nametoindex(interface):
-    """Call the C function if_nametoindex(), or equivalent
-    
-    Note: This function cannot accept a unicode string."""
-    global if_nametoindex
-    try:
-        if_nametoindex = (ctypes.cdll.LoadLibrary
-                          (ctypes.util.find_library("c"))
-                          .if_nametoindex)
-    except (OSError, AttributeError):
-        logger.warning("Doing if_nametoindex the hard way")
-        def if_nametoindex(interface):
-            "Get an interface index the hard way, i.e. using fcntl()"
-            SIOCGIFINDEX = 0x8933  # From /usr/include/linux/sockios.h
-            with contextlib.closing(socket.socket()) as s:
-                ifreq = fcntl.ioctl(s, SIOCGIFINDEX,
-                                    struct.pack(str("16s16x"),
-                                                interface))
-            interface_index = struct.unpack(str("I"),
-                                            ifreq[16:20])[0]
-            return interface_index
-    return if_nametoindex(interface)
-
-
 def daemon(nochdir = False, noclose = False):
     """See daemon(3).  Standard BSD Unix function.
     
@@ -1851,6 +1970,10 @@
                         " system bus interface")
     parser.add_argument("--no-ipv6", action="store_false",
                         dest="use_ipv6", help="Do not use IPv6")
+    parser.add_argument("--no-restore", action="store_false",
+                        dest="restore", help="Do not restore stored"
+                        " state", default=True)
+    
     options = parser.parse_args()
     
     if options.check:
@@ -1891,7 +2014,7 @@
     # options, if set.
     for option in ("interface", "address", "port", "debug",
                    "priority", "servicename", "configdir",
-                   "use_dbus", "use_ipv6", "debuglevel"):
+                   "use_dbus", "use_ipv6", "debuglevel", "restore"):
         value = getattr(options, option)
         if value is not None:
             server_settings[option] = value
@@ -1910,6 +2033,15 @@
     use_dbus = server_settings["use_dbus"]
     use_ipv6 = server_settings["use_ipv6"]
     
+    if debug:
+        initlogger(logging.DEBUG)
+    else:
+        if not debuglevel:
+            initlogger()
+        else:
+            level = getattr(logging, debuglevel.upper())
+            initlogger(level)
+    
     if server_settings["servicename"] != "Mandos":
         syslogger.setFormatter(logging.Formatter
                                ('Mandos (%s) [%%(process)d]:'
@@ -1969,14 +2101,6 @@
         if error[0] != errno.EPERM:
             raise error
     
-    if not debug and not debuglevel:
-        syslogger.setLevel(logging.WARNING)
-        console.setLevel(logging.WARNING)
-    if debuglevel:
-        level = getattr(logging, debuglevel.upper())
-        syslogger.setLevel(level)
-        console.setLevel(level)
-    
     if debug:
         # Enable all possible GnuTLS debugging
         
@@ -2024,9 +2148,10 @@
             server_settings["use_dbus"] = False
             tcp_server.use_dbus = False
     protocol = avahi.PROTO_INET6 if use_ipv6 else avahi.PROTO_INET
-    service = AvahiService(name = server_settings["servicename"],
-                           servicetype = "_mandos._tcp",
-                           protocol = protocol, bus = bus)
+    service = AvahiServiceToSyslog(name =
+                                   server_settings["servicename"],
+                                   servicetype = "_mandos._tcp",
+                                   protocol = protocol, bus = bus)
     if server_settings["interface"]:
         service.interface = (if_nametoindex
                              (str(server_settings["interface"])))
@@ -2038,23 +2163,128 @@
     if use_dbus:
         client_class = functools.partial(ClientDBusTransitional,
                                          bus = bus)
-    def client_config_items(config, section):
-        special_settings = {
-            "approved_by_default":
-                lambda: config.getboolean(section,
-                                          "approved_by_default"),
-            }
-        for name, value in config.items(section):
+    
+    special_settings = {
+        # Some settings need to be accessd by special methods;
+        # booleans need .getboolean(), etc.  Here is a list of them:
+        "approved_by_default":
+            lambda section:
+            client_config.getboolean(section, "approved_by_default"),
+        }
+    # Construct a new dict of client settings of this form:
+    # { client_name: {setting_name: value, ...}, ...}
+    # with exceptions for any special settings as defined above
+    client_settings = dict((clientname,
+                           dict((setting,
+                                 (value
+                                  if setting not in special_settings
+                                  else special_settings[setting]
+                                  (clientname)))
+                                for setting, value in
+                                client_config.items(clientname)))
+                          for clientname in client_config.sections())
+    
+    old_client_settings = {}
+    clients_data = []
+    
+    # Get client data and settings from last running state.
+    if server_settings["restore"]:
+        try:
+            with open(stored_state_path, "rb") as stored_state:
+                clients_data, old_client_settings = (pickle.load
+                                                     (stored_state))
+            os.remove(stored_state_path)
+        except IOError as e:
+            logger.warning("Could not load persistent state: {0}"
+                           .format(e))
+            if e.errno != errno.ENOENT:
+                raise
+    
+    with Crypto() as crypt:
+        for client in clients_data:
+            client_name = client["name"]
+            
+            # Decide which value to use after restoring saved state.
+            # We have three different values: Old config file,
+            # new config file, and saved state.
+            # New config value takes precedence if it differs from old
+            # config value, otherwise use saved state.
+            for name, value in client_settings[client_name].items():
+                try:
+                    # For each value in new config, check if it
+                    # differs from the old config value (Except for
+                    # the "secret" attribute)
+                    if (name != "secret" and
+                        value != old_client_settings[client_name]
+                        [name]):
+                        setattr(client, name, value)
+                except KeyError:
+                    pass
+            
+            # Clients who has passed its expire date can still be
+            # enabled if its last checker was sucessful.  Clients
+            # whose checker failed before we stored its state is
+            # assumed to have failed all checkers during downtime.
+            if client["enabled"] and client["last_checked_ok"]:
+                if ((datetime.datetime.utcnow()
+                     - client["last_checked_ok"])
+                    > client["interval"]):
+                    if client["last_checker_status"] != 0:
+                        client["enabled"] = False
+                    else:
+                        client["expires"] = (datetime.datetime
+                                             .utcnow()
+                                             + client["timeout"])
+            
+            client["changedstate"] = (multiprocessing_manager
+                                      .Condition
+                                      (multiprocessing_manager
+                                       .Lock()))
+            if use_dbus:
+                new_client = (ClientDBusTransitional.__new__
+                              (ClientDBusTransitional))
+                tcp_server.clients[client_name] = new_client
+                new_client.bus = bus
+                for name, value in client.iteritems():
+                    setattr(new_client, name, value)
+                client_object_name = unicode(client_name).translate(
+                    {ord("."): ord("_"),
+                     ord("-"): ord("_")})
+                new_client.dbus_object_path = (dbus.ObjectPath
+                                               ("/clients/"
+                                                + client_object_name))
+                DBusObjectWithProperties.__init__(new_client,
+                                                  new_client.bus,
+                                                  new_client
+                                                  .dbus_object_path)
+            else:
+                tcp_server.clients[client_name] = (Client.__new__
+                                                   (Client))
+                for name, value in client.iteritems():
+                    setattr(tcp_server.clients[client_name],
+                            name, value)
+            
             try:
-                yield (name, special_settings[name]())
-            except KeyError:
-                yield (name, value)
-    
-    tcp_server.clients.update(set(
-            client_class(name = section,
-                         config= dict(client_config_items(
-                        client_config, section)))
-            for section in client_config.sections()))
+                tcp_server.clients[client_name].secret = (
+                    crypt.decrypt(tcp_server.clients[client_name]
+                                  .encrypted_secret,
+                                  client_settings[client_name]
+                                  ["secret"]))
+            except CryptoError:
+                # If decryption fails, we use secret from new settings
+                tcp_server.clients[client_name].secret = (
+                    client_settings[client_name]["secret"])
+    
+    # Create/remove clients based on new changes made to config
+    for clientname in set(old_client_settings) - set(client_settings):
+        del tcp_server.clients[clientname]
+    for clientname in set(client_settings) - set(old_client_settings):
+        tcp_server.clients[clientname] = (client_class(name
+                                                       = clientname,
+                                                       config =
+                                                       client_settings
+                                                       [clientname]))
+    
     if not tcp_server.clients:
         logger.warning("No clients defined")
         
@@ -2103,7 +2333,8 @@
             def GetAllClients(self):
                 "D-Bus method"
                 return dbus.Array(c.dbus_object_path
-                                  for c in tcp_server.clients)
+                                  for c in
+                                  tcp_server.clients.itervalues())
             
             @dbus.service.method(_interface,
                                  out_signature="a{oa{sv}}")
@@ -2111,15 +2342,15 @@
                 "D-Bus method"
                 return dbus.Dictionary(
                     ((c.dbus_object_path, c.GetAll(""))
-                     for c in tcp_server.clients),
+                     for c in tcp_server.clients.itervalues()),
                     signature="oa{sv}")
             
             @dbus.service.method(_interface, in_signature="o")
             def RemoveClient(self, object_path):
                 "D-Bus method"
-                for c in tcp_server.clients:
+                for c in tcp_server.clients.itervalues():
                     if c.dbus_object_path == object_path:
-                        tcp_server.clients.remove(c)
+                        del tcp_server.clients[c.name]
                         c.remove_from_connection()
                         # Don't signal anything except ClientRemoved
                         c.disable(quiet=True)
@@ -2139,11 +2370,52 @@
         service.cleanup()
         
         multiprocessing.active_children()
+        if not (tcp_server.clients or client_settings):
+            return
+        
+        # Store client before exiting. Secrets are encrypted with key
+        # based on what config file has. If config file is
+        # removed/edited, old secret will thus be unrecovable.
+        clients = []
+        with Crypto() as crypt:
+            for client in tcp_server.clients.itervalues():
+                key = client_settings[client.name]["secret"]
+                client.encrypted_secret = crypt.encrypt(client.secret,
+                                                        key)
+                client_dict = {}
+                
+                # A list of attributes that will not be stored when
+                # shutting down.
+                exclude = set(("bus", "changedstate", "secret"))
+                for name, typ in (inspect.getmembers
+                                  (dbus.service.Object)):
+                    exclude.add(name)
+                
+                client_dict["encrypted_secret"] = (client
+                                                   .encrypted_secret)
+                for attr in client.client_structure:
+                    if attr not in exclude:
+                        client_dict[attr] = getattr(client, attr)
+                
+                clients.append(client_dict)
+                del client_settings[client.name]["secret"]
+        
+        try:
+            with os.fdopen(os.open(stored_state_path,
+                                   os.O_CREAT|os.O_WRONLY|os.O_TRUNC,
+                                   0600), "wb") as stored_state:
+                pickle.dump((clients, client_settings), stored_state)
+        except (IOError, OSError) as e:
+            logger.warning("Could not save persistent state: {0}"
+                           .format(e))
+            if e.errno not in (errno.ENOENT, errno.EACCES):
+                raise
+        
+        # Delete all clients, and settings from config
         while tcp_server.clients:
-            client = tcp_server.clients.pop()
+            name, client = tcp_server.clients.popitem()
             if use_dbus:
                 client.remove_from_connection()
-            client.disable_hook = None
             # Don't signal anything except ClientRemoved
             client.disable(quiet=True)
             if use_dbus:
@@ -2151,14 +2423,17 @@
                 mandos_dbus_service.ClientRemoved(client
                                                   .dbus_object_path,
                                                   client.name)
+        client_settings.clear()
     
     atexit.register(cleanup)
     
-    for client in tcp_server.clients:
+    for client in tcp_server.clients.itervalues():
         if use_dbus:
             # Emit D-Bus signal
             mandos_dbus_service.ClientAdded(client.dbus_object_path)
-        client.enable()
+        # Need to initiate checking of clients
+        if client.enabled:
+            client.init_checker()
     
     tcp_server.enable()
     tcp_server.server_activate()

=== modified file 'mandos-options.xml'
--- mandos-options.xml	2009-02-13 05:38:21 +0000
+++ mandos-options.xml	2011-11-09 17:16:03 +0000
@@ -85,5 +85,10 @@
     to the server if this option is turned off.  <emphasis>Only
     advanced users should consider changing this option</emphasis>.
   </para>
+
+  <para id="restore">
+    This option controls whether the server will restore any state from
+    last time it ran. Default is to try restore last state.
+  </para>
   
 </section>

=== modified file 'mandos.xml'
--- mandos.xml	2011-10-05 16:00:56 +0000
+++ mandos.xml	2011-11-09 17:16:03 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos">
-<!ENTITY TIMESTAMP "2011-10-03">
+<!ENTITY TIMESTAMP "2011-10-22">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -94,6 +94,8 @@
       <arg><option>--no-dbus</option></arg>
       <sbr/>
       <arg><option>--no-ipv6</option></arg>
+      <sbr/>
+      <arg><option>--no-restore</option></arg>
     </cmdsynopsis>
     <cmdsynopsis>
       <command>&COMMANDNAME;</command>
@@ -275,6 +277,13 @@
 	  <xi:include href="mandos-options.xml" xpointer="ipv6"/>
 	</listitem>
       </varlistentry>
+      
+      <varlistentry>
+	<term><option>--no-restore</option></term>
+	<listitem>
+	  <xi:include href="mandos-options.xml" xpointer="restore"/>
+	</listitem>
+      </varlistentry>
     </variablelist>
   </refsect1>
   
@@ -509,9 +518,6 @@
       Debug mode is conflated with running in the foreground.
     </para>
     <para>
-      The console log messages do not show a time stamp.
-    </para>
-    <para>
       This server does not check the expire time of clients’ OpenPGP
       keys.
     </para>

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2011-11-11 09:10:58 +0000
+++ plugins.d/mandos-client.c	2011-10-05 16:00:56 +0000
@@ -123,7 +123,6 @@
 #define PATHDIR "/conf/conf.d/mandos"
 #define SECKEY "seckey.txt"
 #define PUBKEY "pubkey.txt"
-#define HOOKDIR "/lib/mandos/network-hooks.d"
 
 bool debug = false;
 static const char mandos_protocol_version[] = "1";
@@ -1086,94 +1085,12 @@
   errno = old_errno;
 }
 
-bool get_flags(const char *ifname, struct ifreq *ifr){
-  int ret;
-  
-  int s = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
-  if(s < 0){
-    perror_plus("socket");
-    return false;
-  }
-  strcpy(ifr->ifr_name, ifname);
-  ret = ioctl(s, SIOCGIFFLAGS, ifr);
-  if(ret == -1){
-    if(debug){
-      perror_plus("ioctl SIOCGIFFLAGS");
-    }
-    return false;
-  }
-  return true;
-}
-
-bool good_flags(const char *ifname, const struct ifreq *ifr){
-  
-  /* Reject the loopback device */
-  if(ifr->ifr_flags & IFF_LOOPBACK){
-    if(debug){
-      fprintf(stderr, "Rejecting loopback interface \"%s\"\n",
-	      ifname);
-    }
-    return false;
-  }
-  /* Accept point-to-point devices only if connect_to is specified */
-  if(connect_to != NULL and (ifr->ifr_flags & IFF_POINTOPOINT)){
-    if(debug){
-      fprintf(stderr, "Accepting point-to-point interface \"%s\"\n",
-	      ifname);
-    }
-    return true;
-  }
-  /* Otherwise, reject non-broadcast-capable devices */
-  if(not (ifr->ifr_flags & IFF_BROADCAST)){
-    if(debug){
-      fprintf(stderr, "Rejecting non-broadcast interface \"%s\"\n",
-	      ifname);
-    }
-    return false;
-  }
-  /* Reject non-ARP interfaces (including dummy interfaces) */
-  if(ifr->ifr_flags & IFF_NOARP){
-    if(debug){
-      fprintf(stderr, "Rejecting non-ARP interface \"%s\"\n", ifname);
-    }
-    return false;
-  }
-  
-  /* Accept this device */
-  if(debug){
-    fprintf(stderr, "Interface \"%s\" is good\n", ifname);
-  }
-  return true;
-}
-
 /* 
  * This function determines if a directory entry in /sys/class/net
  * corresponds to an acceptable network device.
  * (This function is passed to scandir(3) as a filter function.)
  */
 int good_interface(const struct dirent *if_entry){
-  int ret;
-  if(if_entry->d_name[0] == '.'){
-    return 0;
-  }
-  struct ifreq ifr;
-
-  if(not get_flags(if_entry->d_name, &ifr)){
-    return 0;
-  }
-  
-  if(not good_flags(if_entry->d_name, &ifr)){
-    return 0;
-  }
-  return 1;
-}
-
-/* 
- * This function determines if a directory entry in /sys/class/net
- * corresponds to an acceptable network device which is up.
- * (This function is passed to scandir(3) as a filter function.)
- */
-int up_interface(const struct dirent *if_entry){
   ssize_t ssret;
   char *flagname = NULL;
   if(if_entry->d_name[0] == '.'){
@@ -1241,12 +1158,6 @@
     }
     return 0;
   }
-
-  /* Reject down interfaces */
-  if(not (flags & IFF_UP)){
-    return 0;
-  }
-  
   /* Accept point-to-point devices only if connect_to is specified */
   if(connect_to != NULL and (flags & IFF_POINTOPOINT)){
     if(debug){
@@ -1290,50 +1201,6 @@
   return 1;
 }
 
-/* Is this directory entry a runnable program? */
-int runnable_hook(const struct dirent *direntry){
-  int ret;
-  struct stat st;
-  
-  if((direntry->d_name)[0] == '\0'){
-    /* Empty name? */
-    return 0;
-  }
-  
-  /* Save pointer to last character */
-  char *end = strchr(direntry->d_name, '\0')-1;
-  
-  if(*end == '~'){
-    /* Backup name~ */
-    return 0;
-  }
-  
-  if(((direntry->d_name)[0] == '#')
-     and (*end == '#')){
-    /* Temporary #name# */
-    return 0;
-  }
-  
-  /* XXX more rules here */
-  
-  ret = stat(direntry->d_name, &st);
-  if(ret == -1){
-    if(debug){
-      perror_plus("Could not stat plugin");
-    }
-    return 0;
-  }
-  if(not (st.st_mode & S_ISREG)){
-    /* Not a regular file */
-    return 0;
-  }
-  if(not (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))){
-    /* Not executable */
-    return 0;
-  }
-  return 1;
-}
-
 int avahi_loop_with_timeout(AvahiSimplePoll *s, int retry_interval){
   int ret;
   struct timespec now;
@@ -1651,47 +1518,10 @@
     }
   }
   
-  /* Find network hooks and run them */
-  {
-    struct dirent **direntries;
-    struct dirent *direntry;
-    int numhooks = scandir(HOOKDIR, &direntries, runnable_hook,
-			   alphasort);
-    int devnull = open("/dev/null", O_RDONLY);
-    for(int i = 0; i < numhooks; i++){
-      direntry = direntries[0];
-      char *fullname = NULL;
-      ret = asprintf(&fullname, "%s/%s", tempdir,
-		     direntry->d_name);
-      if(ret < 0){
-	perror_plus("asprintf");
-	continue;
-      }
-      pid_t hook_pid = fork();
-      if(hook_pid == 0){
-	/* Child */
-	dup2(devnull, STDIN_FILENO);
-	close(devnull);
-	dup2(STDERR_FILENO, STDOUT_FILENO);
-	setenv("DEVICE", interface, 1);
-	setenv("VERBOSE", debug ? "1" : "0", 1);
-	setenv("MODE", "start", 1);
-	/* setenv( XXX more here */
-	ret = execl(fullname, direntry->d_name, "start");
-	perror_plus("execl");
-      }
-      free(fullname);
-      if(quit_now){
-	goto end;
-      }
-    }
-    close(devnull);
-  }
-  
   if(not debug){
     avahi_set_log_function(empty_log);
   }
-  
+
   if(interface[0] == '\0'){
     struct dirent **direntries;
     ret = scandir(sys_class_net, &direntries, good_interface,
@@ -2039,7 +1869,7 @@
     if (not quit_now){
       exitcode = EXIT_SUCCESS;
     }
-    
+
     goto end;
   }
   
@@ -2142,8 +1972,6 @@
     }
   }
   
-  /* XXX run network hooks "stop" here  */
-  
   /* Take down the network interface */
   if(take_down_interface){
     /* Re-raise priviliges */

